;; Generated package description from boxy-headings.el  -*- no-byte-compile: t -*-
(define-package "boxy-headings" "2.1.8" "View org files in a boxy diagram" '((emacs "26.1") (boxy "2.0") (org "9.4")) :commit "090aee55d8d613b9dee732d787e155d3bf8eb5cd" :authors '(("Amy Grinn" . "grinn.amy@gmail.com")) :maintainer '("Amy Grinn" . "grinn.amy@gmail.com") :keywords '("tools") :url "https://gitlab.com/grinn.amy/boxy-headings")
